from flask import g, Flask, jsonify
import re
import sqlite3
import os

app = Flask(__name__)
DATABASE = './database.db'
all_tokens = {}

def get_db():
    """
    Establishes a new SQLite database connection if one does not already exist in the application context.
    """
    if 'db' not in g:
        g.db = sqlite3.connect(DATABASE, timeout=10)
        g.db.row_factory = sqlite3.Row
    return g.db

def init_db():
    """
    Initializes the SQLite database using the provided schema.sql file.
    Executes the SQL schema script to create necessary tables.
    """
    with app.app_context():
        db = get_db()
        try: 
            with open('schema.sql', 'r') as f:
                sql_script = f.read()
            db.executescript(sql_script)
            db.commit()
        except Exception as e:
            print("failed to init db", e)

def add_user(email, password, firstname, familyName, gender, city, country):
    """
    Inserts a new user into the users table.
    """

    with app.app_context():
        db = get_db()
        cursor = db.cursor()
        cursor.execute("INSERT INTO users (email, password, firstname, familyName, gender, city, country) VALUES (?, ?, ?, ?, ?, ?, ?)", (email, password, firstname, familyName, gender, city, country))
        db.commit()

def validate_email(email):
    if email is None:
        return True
    email_regex = r'^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$'
    return re.match(email_regex, email) is not None

def validate_password(password):
    return password is not None and len(password) >= 6

def validate_name(name):
    return name is not None and name.isalpha()

def validate_gender(gender):
    return gender is not None and gender.lower() in ['male', 'female', 'others']

def validate_city_or_country(region):
    return region is not None 

def validate_inputs(email, password, firstname, familyName, gender, city, country):
    errors = {}

    if not validate_email(email):
        errors['email'] = "Invalid email format."
    
    if not validate_password(password):
        errors['password'] = "Password must be at least 6 characters long."
    
    if not validate_name(firstname):
        errors['firstname'] = "Firstname should contain only alphabetic characters."
    
    if not validate_name(familyName):
        errors['familyName'] = "Family name should contain only alphabetic characters."
    
    if not validate_gender(gender):
        errors['gender'] = "Gender must be 'Male' or 'Female' or 'Others'."
    
    if not validate_city_or_country(city):
        errors['city'] = "City should contain only alphabetic characters."
    
    if not validate_city_or_country(country):
        errors['country'] = "Country should contain only alphabetic characters."
    return errors

def is_email_exist( email : str ):
    with app.app_context():
        db = get_db()
        cursor = db.cursor()
        return cursor.execute('SELECT * FROM users WHERE email=?',(email,)).fetchone()

def check_sign_in_params( username: str, password: str ):
    with app.app_context():
        db = get_db()
        cursor = db.cursor()
        return cursor.execute('SELECT * FROM users WHERE email=? AND password=?', (username, password)).fetchone()

def change_password( email : str , oldpassword: str, newpassword: str):
    with app.app_context():
        db = get_db()
        cursor = db.cursor()

        cursor.execute("SELECT password FROM users WHERE email=?", (email,))

        result = cursor.fetchone()

        if not result or result[0] != oldpassword:
            return jsonify({"success": False, "message": "Old password is incorrect", "data": None}), 401
        
        try:
            cursor.execute("UPDATE users SET password=? WHERE email=?", (newpassword, email))
            db.commit()

            return jsonify({"success": True, "message": "Password changed successfully", "data": newpassword}), 200
        
        except Exception as e:
            db.rollback()
            return jsonify({"success": False, "message":  "Internal server error", "data": None }), 500
        

def get_user_messages_by_email( email : str ):
    db = get_db()

    #db.row_factory = sqlite3.Row
    cursor = db.cursor()

    result = cursor.execute('SELECT id, writer, messages, latitude, longitude FROM users_messages WHERE email=?', (email,)).fetchall()

    if not result:
        return jsonify({"success": False, "message" : "Not found", "data" : None}), 404
    
    messages = [dict(row) for row in result]
    return jsonify({"success": True, "message" : "Successfully fetched data", "data" : messages }), 200

def get_user_data_by_email( email : str ):
    db = get_db()
    cursor = db.cursor()
    
    result = cursor.execute('SELECT * FROM users WHERE email=?', (email,)).fetchone()

    if not result:
        return jsonify({"success": False, "message" : "Not found", "data" : None}), 404
    
    return jsonify({"success": True, "message" : "Successfully fetched data", "data" : dict(result)}),200

def save_media(email, file_path,file_type):
    db= get_db()
    cursor= db.cursor()

    cursor.execute("INSERT INTO user_media (email, file_path, file_type) VALUES (?, ?, ?)", 
                   (email, file_path, file_type))
    db.commit()
def save_profile(email, file_path):
    db= get_db()
    cursor = db.cursor()

    cursor.execute("UPDATE users SET profile_image = ? WHERE email = ?", (file_path, email))
    db.commit()

def get_user_media(email):
    db = get_db()
    cursor = db.cursor()
    result = cursor.execute("SELECT file_path, file_type FROM user_media WHERE email=?", (email,)).fetchall()
    if not result:
        return jsonify({"success": True, "message" : "No data", "data" : []}), 404
    return jsonify({"success": True, "message": "Success", "data" : [dict(row) for row in result]}), 200


def get_user_profile(email):
    db = get_db()
    cursor = db.cursor()
    result = cursor.execute('SELECT profile_image FROM users WHERE email=?', (email,)).fetchone()
    print(result)
    if not result: 
        return jsonify({"success": False, "message" : "Not found", "data" : None}), 404
    return jsonify({"success": True, "message": "Success", "data" :dict(result)}), 200


def post_messages (email, message, writer, latitude, longitude):
    db = get_db()
    cursor = db.cursor()
    try:
            cursor.execute("INSERT INTO users_messages (email, messages, writer, latitude, longitude) VALUES (?, ?, ?, ?, ?)", (email, message, writer, latitude, longitude))
            db.commit()
            return jsonify({"success": True, "message": "Message posted successfully", "data": message}), 201
        
    except Exception as e:
            db.rollback()
            print("exception", e)
            return jsonify({"success": False, "message": "Internal server error", "data": None}),500
