"""
Flask Backend Server

This server provides authentication, user messaging, profile image uploads,
and media management using Flask, WebSocket, and SQLite.

Endpoints:
- User Auth: /sign_in, /sign_up, /sign_out
- User Data: /get_user_data_by_token, /get_user_data_by_email/<email>
- Messages: /post_message, /get_user_messages_by_token
- Media: /upload_file, /get_user_media, /upload_profile
"""


from database_helper import all_tokens, get_db
from flask import Flask, g, request, render_template, session, jsonify
import uuid
import database_helper
from flask_sock import Sock
from flask_cors import CORS
import os
from werkzeug.utils import secure_filename

app = Flask(__name__)
app.secret_key = "111"
sock = Sock(app)
active_connections = {}
DATABASE = './database.db'
UPLOAD_FOLDER = './media_uploads'
ALLOWED_EXTENSIONS = {'png', 'jpg', 'jpeg', 'gif', 'mp4', 'mp3', 'wav'}
CORS(app, supports_credentials=True)

app.config['UPLOAD_FOLDER'] = UPLOAD_FOLDER
if not os.path.exists(UPLOAD_FOLDER):
    os.makedirs(UPLOAD_FOLDER)

@app.route('/')
def home():
    """
    Renders the main client HTML page for the frontend application.
    """
    return render_template('client.html')

@app.teardown_appcontext
def close_connection(exception):
    """
    Closes the database connection at the end of the request lifecycle.
    """
    db = g.pop('db', None)
    print(exception)
    if db is not None:
        db.close()

@sock.route('/ws')
def websocket_connection(ws):
    """
    Manages WebSocket connections, stores them, and handles disconnections.
    """

    token = request.headers.get("Authorization")

    user_id = session.get("user_id")

    if not token or not user_id:
        ws.close()
        return

    # Store the WebSocket connection for this user
    active_connections[user_id] = ws

    try:
        while True:
            message = ws.receive()
            if message == "ping":
                ws.send("pong")  # Keep the connection alive
    except:
        pass
    finally:
        # Remove user from active connections when disconnected
        if user_id in active_connections:
            del active_connections[user_id]
            

@app.put('/change_password')
def change_password():
    """
    Handles password change requests by validating the old and new passwords.
    """
    try:
        if request.method != "PUT":
             return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        data = request.get_json()
        token = request.headers.get("Authorization")
    
        oldpassword = data.get("oldPassword")
        newpassword = data.get("newPassword")


        if not token or token not in all_tokens:
            return jsonify({"success": False, "message": "Invalid token", "data": None}), 401
        if not oldpassword:
            return jsonify({"success": False, "message": "Old password is required", "data": None}), 400
        if not newpassword:
            return jsonify({"success": False, "message": "New password is required", "data": None}), 400
        if oldpassword == newpassword:
            return jsonify({"success": False, "message": "Old password and new password are same", "data": None}), 400
           
        email = all_tokens[token]
        
        return database_helper.change_password(email, oldpassword, newpassword)
    
    except Exception as e:
        print("Error in sign_in", e)
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500


# have to change method more abstract and specify in the client side
@app.post('/sign_in')
def sign_in():
    """
    Authenticates a user and starts a new session using a token.
    """
    try:
        if request.method != "POST":
             return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        data = request.get_json()

        if not data or "username" not in data or "password" not in data:
            return jsonify({"success": False, "message":  "No data received", "data": None}), 400
        
        email = data.get("username")
        password = data.get("password")

        if not email:
            return jsonify({"success": False, "message": "Email not found", "data": None}), 400
        if not password: 
            return jsonify({"success": False, "message": "Password not found", "data": None}), 400

        if database_helper.check_sign_in_params(email, password):
            
            token = uuid.uuid4().hex
            all_tokens[token] = email

            session["user_id"] = email
            session["session_id"] = token
            
            if email in active_connections:
                old_ws = active_connections[email]
                try: 
                    sign_out()
                    old_ws.send("logout")
                    old_ws.close()
                except:
                    pass
                del active_connections[email]

            return jsonify({ "success": True, "message" : "Successfully signed in", "data" : token }), 200
        return jsonify({ "success": False, "message" : "Failed to sign in", "data" : None }), 401
    
    except Exception as e:
        print("Error in sign_in", e)
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500

@app.post('/sign_up')
def sign_up():
    """
    Registers a new user after validating input and checking for duplicate email.
    """
    try:
        if request.method != "POST":
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        
        data = request.get_json()

        if not data:
            return jsonify({"success": False, "message": "Request body is missing", "data": None}), 400
        
        email = data.get("email")
        password = data.get("password")
        firstname = data.get("firstname")
        familyName = data.get("familyname")
        gender = data.get("gender")
        city = data.get("city")
        country = data.get("country")
        
        # if its not valid parameter
        if database_helper.validate_inputs(email, password, firstname, familyName, gender, city, country):
        
            return jsonify({ "success": False, "message" : "Failed to sign up", "data" :  None }), 400
        
        if database_helper.is_email_exist(email):
            return jsonify({"success": False, "message": "Email already exists", "data": None}), 409
                                            
        database_helper.add_user(email, password, firstname, familyName, gender, city, country)
    
        return jsonify({ "success": True, "message" : "Successfully signed up", "data" : data }), 201
    
    except Exception as e:
        print("Error in sign_up", e)
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500
    

@app.delete('/sign_out')
def sign_out():
    """
    Signs out the current user and removes the session and WebSocket."""
    try: 
        if request.method != 'DELETE':
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405

        token = request.headers.get("Authorization")

        if not token:
            return jsonify({"success": False, "message": "Not found", "data": None}), 400
        
        if token not in all_tokens:
            return jsonify({"success": False, "message": "Invalid token", "data": None}), 401

        email = all_tokens.pop(token, None)  # Remove token from active sessions
        session.clear()  # Clear session data

        # Remove WebSocket connection if exists
        if email in active_connections:
            del active_connections[email]

        return jsonify({"success": True, "message": "Signed out successfully", "data": None}), 200
    
    except Exception as e:
        print(f"Error in sign_out: {e}")  # Log error for debugging
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500

 

@app.get( '/get_user_data_by_token' )
def get_user_data_by_token():
    """
    Fetches user data using a valid token from the header.
    """
    try: 
        if request.method != 'GET':
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        token = request.headers.get("Authorization")

        if not token or token not in all_tokens:
            return jsonify({ "success": False, "message" : "Invalid token", "data" : None }), 401
        
        email = all_tokens[token]
        return database_helper.get_user_data_by_email(email)
    
    except Exception as e:
        print(f"Error in sign_out: {e}")  # Log error for debugging
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500
    
 
@app.get('/get_user_messages_by_email/<email>')
def get_user_messages_by_email (email : str):
    """
    Retrieves messages posted to a specific user by email.
    """
    try: 
        if request.method != 'GET':
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        token = request.headers.get("Authorization")

        if token not in all_tokens:
            return jsonify({"success": False, "message" : "Invalid token", "data" : None}), 401
        
        
        return database_helper.get_user_messages_by_email(email)
    except Exception as e:
        print(f"Error in sign_out: {e}")  # Log error for debugging
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500
         

@app.get('/get_user_messages_by_token')
def get_user_messages_by_token ():
    """
    Retrieves messages for the logged-in user identified by token.
    """
    try: 
        if request.method != 'GET':
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        token = request.headers.get("Authorization")

        if token not in all_tokens:
            return jsonify({"success": False, "message" : "Invalid token", "data" : None}), 401
        
        email = all_tokens[token]
        return database_helper.get_user_messages_by_email(email)
    
    except Exception as e: 
        print(f"Error in sign_out: {e}")  # Log error for debugging
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500
    
@app.get('/get_user_data_by_email/<email>')    
def get_user_data_by_email(email : str):
    """
    Fetches public profile data of a user by their email address.
    """
    try: 
        if request.method != 'GET':
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        
        if email is None:
            return jsonify({"success": False, "message" : "Email is not provided", "data" : None}), 400
        
        token = request.headers.get("Authorization")

        if not token or token not in all_tokens:
            return jsonify({"success": False, "message" : "Invalid token", "data" : None}), 401
        
        result = database_helper.get_user_data_by_email(email)

        if not result:
            return jsonify({"success": False, "message" : "Not found", "data" : None}), 404 
        
        return jsonify({"success": True, "message" : "Successfully fetched data", "data" : dict(result)}), 200
    
    except Exception as e:
        print(f"Error in sign_out: {e}")  # Log error for debugging
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500

@app.post('/post_message')
def post_message():
    """
    Posts a message from the current user to another user's wall with optional location."""
    try: 
        if request.method != 'POST':
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405

        data = request.get_json()

        token = request.headers.get("Authorization")

        email = data.get("email")
        writer = all_tokens[token]
        message = data.get("message")
        latitude = data.get("latitude")
        longitude = data.get("longitude")

        if not token or token not in all_tokens:
            return jsonify({"success": False, "message": "Invalid token", "data": None}), 401
        if not message:
            return jsonify({"success": False, "message": "Not found", "data": None}), 400
        if not email:
            email = all_tokens[token]

        return database_helper.post_messages(email, message, writer, latitude, longitude)

    except Exception as e:
        print("exception", e)
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500
    

def allowed_file(filename: str):
    return '.' in filename and filename.rsplit('.', 1)[1].lower() in ALLOWED_EXTENSIONS

@app.post('/upload_file')
def upload_file():
    """
    Handles media file uploads and stores file info in the database.
    """
    try: 
        if request.method != 'POST':
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        
        token = request.headers.get("Authorization")

        if not token or token not in all_tokens:
            return jsonify({"success": False, "message": "Invalid token"}), 401
        email = all_tokens[token]

        if 'file' not in request.files:
            return jsonify({"success": False, "message": "No file part"}), 400
        
        file = request.files['file']

        if file.filename == '':
            return jsonify({"success": False, "message": "No file part"}), 400
        
        if file and allowed_file(file.filename):
            filename = secure_filename(file.filename)
            file_path = os.path.join(app.config['UPLOAD_FOLDER'], filename)
            file.save(file_path)

            database_helper.save_media(email,file_path, filename.split('.',1)[1].lower())

            return jsonify({"success": True, "message": "File uploaded successfully"}), 201
        
        return jsonify({"success": False, "message": "File type not allowed"}), 400
    
    except Exception as e:
        print("exception", e)
        return jsonify({"success": False, "message": "Internal server error", "data": None}),500
    

@app.post('/upload_profile')
def upload_profile():
    """
    Handles profile image upload and updates user's profile_image path.
    """
    try: 
        if request.method != 'POST':
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        
        token = request.headers.get("Authorization")

        if not token or token not in all_tokens:
            return jsonify({"success": False, "message": "Invalid token"}), 401
        email = all_tokens[token]

        if 'file' not in request.files:
            return jsonify({"success": False, "message": "No file part"}), 400
        
        file = request.files['file']

        if file.filename == '':
            return jsonify({"success": False, "message": "No file part"}), 400
        
        if file and allowed_file(file.filename):
            filename = secure_filename(file.filename)
            file_path = os.path.join(app.config['UPLOAD_FOLDER'], filename)
            file.save(file_path)

            database_helper.save_profile(email,file_path)

            return jsonify({"success": True, "message": "File uploaded successfully"}), 201
        
        return jsonify({"success": False, "message": "File type not allowed"}), 400
    
    except Exception as e:
        print("exception", e)
        return jsonify({"success": False, "message": "Internal server error", "data": None}),500
    
@app.get('/get_user_media')
def get_user_media():
    """
    Retrieves all uploaded media files for the logged-in user.
    """
    try: 
        if request.method != 'GET':
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        token = request.headers.get('Authorization')

        if not token or token not in all_tokens:
            return jsonify({"success": False, "data": None, "message" :"Invalid token"}), 401
        email = all_tokens[token]
        return database_helper.get_user_media(email)

    except Exception as e:
        print("exception", e)
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500

@app.get('/get_user_profile')
def get_user_profile():
    """
    Retrieves the profile image path for the logged-in user.
    """
    try: 
        if request.method != 'GET':
            return jsonify({"success": False, "message": "Method Not Allowed", "data": None}), 405
        token = request.headers.get('Authorization')    

        if not token or token not in all_tokens:
            return jsonify({"success": False, "data": None, "message" :"Invalid token"}), 401
        email = all_tokens[token]
        return database_helper.get_user_profile(email)

    except Exception as e:
        print("exception", e)
        return jsonify({"success": False, "message": "Internal server error", "data": None}), 500


database_helper.init_db()

if __name__ == "__main__":
    database_helper.init_db()
    print("Starting Flask server...")
    app.run(host="127.0.0.1", port=5000, debug=True)

