/**
 * Executes when the window has fully loaded.
 * Determines whether a user is logged in by checking for a token in localStorage.
 * If a token exists, it loads the profile view, user data, and messages.
 * Otherwise, it displays the welcome view.
 * Also, sets up event listeners for form submissions.
 * This is overloading. Overloading and Overriding is different!
 */
const sign_in_message = {
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "No data received": "Please enter both email and password.",
  "Email not found": "Email is missing. Please double-check your input.",
  "Password not found": "Password is missing. Please enter your password.",
  "Failed to sign in": "Failed to sign in. Try it again",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
};

const sign_up_message = {
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "Request body is missing": "Missing user information. Please fill out all required fields.",
  "Failed to sign up": "Signup failed. Please try again.",
  "Email already exists": "This email is already registered. Please use a different one.",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
};

const sign_out_message = {
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "Invalid token": "You are already logged out. Enter the web again",
  "Not found": "Token is not provided. Try it again!",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
};

const get_user_data_by_token_message = {
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "Invalid token": "Failed to fetch user data by the provided token",
  "Not found": "Failed to fetch user data by the provided email",
  "Internal server error": "Server is not working well. Try to logout again",
};
const get_user_data_by_email_message = {
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "Invalid token": "Failed to fetch user data by the provided token",
  "Not found": "There is no data",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
  "Email is not provided": "Email is not provided. Check the login status again!",
};

const get_user_messages_by_email_message = {
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "Invalid token": "Failed to fetch user data by the provided token",
  "Not found": "No user exist. Try it again!",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
};
const get_user_messages_by_token_message = {
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "Invalid token": "Failed to fetch user data by the provided token",
  "Not found": "Failed to fetch messages by the provided token",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
};

const post_messages_message = {
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "Invalid token": "Failed to fetch user data by the provided token",
  "Not found": "Email and message are required",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
};

const change_password_message = {
  "Invalid token": "Failed to fetch user data by the provided token",
  "Old password is required": "Old password is not provided. Try it again!",
  "New password is required": "New password is not provided. Try it again!",
  "Old password and new password are same": "New password cannot be the same as the old password",
  "Old password is incorrect": "Old passwords is incorrect. Please check the password",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
  "Invalid password": "Password should be longer than 6 letters. Try it again!",
};

const upload_file_message = {
  "Invalid token": "Failed to fetch user data by the provided token",
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
  "No file part": "File is not uploaded. Try it again!",
  "File type not allowed": "File type should be one of 'png', 'jpg', 'jpeg', 'gif', 'mp4', 'mp3', 'wav'.",
};
const upload_profile_message = {
  "Invalid token": "Failed to fetch user data by the provided token",
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
  "No file part": "File is not uploaded. Try it again!",
  "File type not allowed": "File type should be one of 'png', 'jpg', 'jpeg'",
};

const get_file_message = {
  "Invalid token": "Failed to fetch user data by the provided token",
  "Method Not Allowed": "Inappropriate http request. Try it again!",
  "Internal server error": "Oops! Something went wrong on our end. Please try logging in again later.",
  "Not found": "No media here.",
};

let ws;
let sessionId;
const BASE_URL = "http://127.0.0.1:5000";

window.onload = async function () {
  //get token from localstorage. It may or may not exist.
  const token = localStorage.getItem("token");

  //If there is a token, your session is not expired, so you don't have to login again.
  if (token) {
    displayView("profileView");
    loadProfileData(token);
    loadUserMedia();
    loadMessages(token);
  } else {
    displayView("welcomeView");
  }
  setupEventListeners();
};

function sendFileRequest(method, url, data = null, token = null, callback) {
  const xhr = new XMLHttpRequest();
  xhr.open(method, url, true);

  // Remove Content-Type, as FormData automatically sets it correctly
  if (token) {
    xhr.setRequestHeader("Authorization", token);
  }

  xhr.onreadystatechange = function () {
    if (xhr.readyState === 4) {
      try {
        const response = JSON.parse(xhr.responseText);
        callback(response);
        response.status = xhr.status;
        if (xhr.status === 201) {
          loadUserMedia();
        } else {
          document.getElementById("mediaUploadmessage").innerText = "failed to upload";
          setTimeout(() => ((document.getElementById("mediaUploadmessage").innerText = ""), 3000));
        }
      } catch (e) {
        console.error("Error parsing JSON response", e);
        document.getElementById("mediaUploadmessage").innerText = "Server error occur. Try it later";
        setTimeout(() => ((document.getElementById("mediaUploadmessage").innerText = ""), 3000));
      }
    }
  };

  xhr.send(data);
}

function sendRequest(method, url, data = null, token = null, callback) {
  const xhr = new XMLHttpRequest();
  xhr.open(method, url, true);
  xhr.setRequestHeader("Content-Type", "application/json");
  if (token) {
    xhr.setRequestHeader("Authorization", token);
  }

  xhr.onreadystatechange = function () {
    if (xhr.readyState === 4) {
      //callback(JSON.parse(xhr.responseText));

      try {
        const response = JSON.parse(xhr.responseText);
        response.status = xhr.status;
        callback(response);
      } catch (e) {
        console.error("Server returned invalid JSON. Check console for details.");
      }
    }
  };

  xhr.send(data ? JSON.stringify(data) : null);
}
/**
 * Updates the main application view by replacing its content with the specified view.
 * Switch between the "welcome" and "profile" views.
 * @param {string} viewId
 */

function displayView(viewId) {
  const appDiv = document.getElementById("bodyView");
  const viewContent = document.getElementById(viewId).innerHTML;
  appDiv.innerHTML = viewContent;
}

/**
 * Manages event listeners for form submissions.
 * Prevents the default form submission behavior and calls the appropriate function.
 */
function setupEventListeners() {
  // Register eventlistener to submit event.
  // If submit events happen, below function is going to work.
  // We check the form's id, so that we can call proper function for each submit event.
  document.addEventListener("submit", function (event) {
    if (event.target.id === "loginForm") {
      // Before call each function, we need to prevent default submitting, because we need to check the value before submitting. (Like validation... etc..)
      event.preventDefault();
      login();
    } else if (event.target.id === "signupForm") {
      event.preventDefault();
      signup();
    } else if (event.target.id === "changePasswordForm") {
      event.preventDefault();
      changePassword();
    } else if (event.target.id === "browseForm") {
      event.preventDefault();
      browseUser();
    }
  });
}
function connectWebSocket(sessionId) {
  if (ws) {
    ws.close(); // Close any existing connection
  }

  ws = new WebSocket(`ws://localhost:5000/ws?session_id=${sessionId}`);

  ws.onmessage = function (event) {
    if (event.data === "logout") {
      alert("You have been logged out because you logged in from another device.");
      window.location.href = "/login"; // Redirect to login page
    }
  };

  ws.onclose = function () {
    setTimeout(() => connectWebSocket(sessionId), 3000); // Auto-reconnect
  };
}
/**
 * Manages login process.
 * Send login credentials to the server and update the UI.
 */
async function login() {
  const email = document.getElementById("loginEmail").value;
  const password = document.getElementById("loginPassword").value;

  sendRequest("POST", `${BASE_URL}/sign_in`, { username: email, password: password }, null, function (result) {
    if (result.status == 200) {
      localStorage.setItem("token", result.data);
      displayView("profileView");
      loadProfileData(result.data);
      loadMessages(result.data);
      connectWebSocket(token);
    } else {
      document.getElementById("loginError").textContent = sign_in_message[result.message];
    }
  });
}

/**
 * Manages signup process.
 * Collects form data and send it to the server.
 */
async function signup() {
  // Makes formData structure and put input values in the structure.
  const formData = {
    email: document.getElementById("signupEmail").value,
    password: document.getElementById("signupPassword").value,
    firstname: document.getElementById("signupFirstName").value,
    familyname: document.getElementById("signupFamilyName").value,
    gender: document.getElementById("signupGender").value,
    city: document.getElementById("signupCity").value,
    country: document.getElementById("signupCountry").value,
  };

  sendRequest("POST", `${BASE_URL}/sign_up`, formData, null, function (result) {
    if (result.status == 201) {
      document.getElementById("signupSuccess").textContent = "Signup Success! Log in please";

      setTimeout(() => displayView("welcomeView"), 5000);
    } else {
      document.getElementById("passwordError").textContent = sign_up_message[result.message];
    }
  });
}

/**
 * Gets user info from server and display it.
 * @param {string} token
 */
async function loadProfileData(token) {
  sendRequest("GET", `${BASE_URL}/get_user_data_by_token`, null, token, function (result) {
    if (result.status == 200) {
      document.getElementById("profileName").textContent = result.data.firstname;
      document.getElementById("profileEmail").textContent = result.data.email;
      document.getElementById("profileGender").textContent = result.data.gender;
      document.getElementById("profileCity").textContent = result.data.city;
      document.getElementById("profileCountry").textContent = result.data.country;
      document.getElementById("profileImageContainer").src = result.data.profile_image
        ? result.data.profile_image
        : "https://cdn.pixabay.com/photo/2012/04/26/19/43/profile-42914_1280.png";
    } else {
      document.getElementById("profileUploadError").innerText = get_user_data_by_token_message[result.message];
    }
  });
}

/**
 * Posts Message to server.
 * If error happens, shows user a response message.
 */
async function postMessage() {
  const token = localStorage.getItem("token");
  const content = document.getElementById("postMessageContent").value;
  if (!navigator.geolocation) {
    alert("Geolocation is not supported.");
    return;
  }
  navigator.geolocation.getCurrentPosition(
    function (position) {
      const data = {
        email: null,
        message: content,
        latitude: position.coords.latitude,
        longitude: position.coords.longitude,
      };
      sendRequest("POST", `${BASE_URL}/post_message`, data, token, function (result) {
        if (result.status == 201) {
          document.getElementById("postMessageContent").value = "";
          loadMessages(token);
        } else {
          document.getElementById("postMessageError").textContent = post_messages_message[result.message];
        }
      });
    },
    function (error) {
      alert("Failed to get location.");
      console.error(error);
    }
  );
}

/**
 * Fetches and displays messages for the logged-in user
 * @param {string} token
 */
async function loadMessages(token) {
  sendRequest("GET", `${BASE_URL}/get_user_messages_by_token`, null, token, async function (result) {
    const wall = document.getElementById("messageWall");
    if (result.success && result.status == 200 && result.data) {
      wall.innerHTML = "";

      for (const msg of result.data) {
        let location = "Location not available";

        if (msg.latitude && msg.longitude) {
          location = await fetchAddress(msg.latitude, msg.longitude);
        }

        const messageHTML = `<p><b>${msg.writer}:</b> ${msg.messages}<br><small>${location}</small></p>`;
        wall.innerHTML += messageHTML;
      }
    }
  });
}

async function refreshMessages() {
  const token = localStorage.getItem("token");
  loadMessages(token);
}
/**
 * Changes the user password.
 */
async function changePassword() {
  const token = localStorage.getItem("token");
  const oldPassword = document.getElementById("oldPassword").value;
  const newPassword = document.getElementById("newPassword").value;
  sendRequest(
    "PUT",
    `${BASE_URL}/change_password`,
    { oldPassword: oldPassword, newPassword: newPassword },
    token,
    function (result) {
      if (result.status == 200) {
        document.getElementById("oldPassword").value = "";
        document.getElementById("newPassword").value = "";
        document.getElementById("changePasswordError").textContent = "changed successfully";
        setTimeout(() => (document.getElementById("changePasswordError").textContent = ""), 3000);
      } else {
        document.getElementById("changePasswordError").textContent = change_password_message[result.message];
        setTimeout(() => (document.getElementById("changePasswordError").textContent = ""), 3000);
      }
    }
  );
}

/**
 * finds the user
 */
async function browseUser() {
  const token = localStorage.getItem("token");
  const email = document.getElementById("browseEmail").value;

  sendRequest("GET", `${BASE_URL}/get_user_data_by_email/${email}`, null, token, function (result) {
    if (result.success && result.status == 200) {
      document.getElementById("browseUserProfile").style.display = "block";
      document.getElementById("browseMessagesSection").style.display = "block";
      document.getElementById("browseName").textContent = result.data.firstname;
      document.getElementById("browseEmailResult").textContent = result.data.email;
      document.getElementById("browseGender").textContent = result.data.gender;
      document.getElementById("browseCity").textContent = result.data.city;
      document.getElementById("browseCountry").textContent = result.data.country;
      document.getElementById("browseError").textContent = "";
      loadMessagesForUser(email);
    } else {
      document.getElementById("browseError").textContent = get_user_messages_by_email_message[result.message];
      document.getElementById("browseUserProfile").style.display = "none";
      document.getElementById("browseMessagesSection").style.display = "none";
    }
  });
}

/**
 * Get messages data from server and shows it in UI.
 * @param {string} email
 */
async function loadMessagesForUser(email) {
  const token = localStorage.getItem("token");
  sendRequest("GET", `${BASE_URL}/get_user_messages_by_email/${email}`, null, token, async function (result) {
    const wall = document.getElementById("browseMessagesSection");

    if (result.success && result.data && result.status == 200) {
      wall.innerHTML = "";

      for (const msg of result.data) {
        let location = "Location not available";

        if (msg.latitude && msg.longitude) {
          location = await fetchAddress(msg.latitude, msg.longitude);
        }

        const messageHTML = `<p><b>${msg.writer}:</b> ${msg.messages}<br><small>${location}</small></p>`;
        wall.innerHTML += messageHTML;
      }
    } else {
      wall.innerHTML = `<p>${get_user_messages_by_email_message[result.message]}</p>`;
    }
  });
}

/**
 * Posts message to user.
 */
async function postMessageToUser() {
  const token = localStorage.getItem("token");
  const email = document.getElementById("browseEmailResult").textContent;
  const content = document.getElementById("browseMessageContent").value;

  sendRequest("POST", `${BASE_URL}/post_message`, { message: content, email: email }, token, function (result) {
    if (result.success && result.status == 201) {
      document.getElementById("browseMessageContent").value = "";
      loadMessagesForUser(email);
    } else {
      document.getElementById("postMessageError").innerText = post_messages_message[result.message];
    }
  });
}

/**
 * Compared two inputs (password input) and if its not same, shows error message.
 */
function checkPasswordMatch() {
  const password = document.getElementById("signupPassword").value;
  const confirmPassword = document.getElementById("signupPasswordConfirm").value;
  const errorMessage = document.getElementById("passwordError");

  if (password !== confirmPassword) {
    errorMessage.textContent = "Passwords do not match!";
  } else {
    errorMessage.textContent = "";
  }
}
/**
 * Manages logout
 */
async function logout() {
  const token = localStorage.getItem("token");

  if (!token) {
    displayView("welcomeView");
    return;
  }

  sendRequest("DELETE", `${BASE_URL}/sign_out`, null, token, function (result) {
    if (result.success && result.status == 200) {
      localStorage.removeItem("token");
      displayView("welcomeView"); // Redirect to welcome page
    } else {
      //should changed to alert in ui
      document.getElementById("logoutError").innerText = sign_out_message[result.message];
      delay(2000);
      displayView("welcomeView");
    }

    if (ws) {
      ws.close(); // Close WebSocket connection
    }
  });
}
async function loadProfileData(token) {
  sendRequest("GET", `${BASE_URL}/get_user_data_by_token`, null, token, function (result) {
    if (result.success && result.status == 200) {
      document.getElementById("profileName").textContent = result.data.firstname;
      document.getElementById("profileEmail").textContent = result.data.email;
      document.getElementById("profileGender").textContent = result.data.gender;
      document.getElementById("profileCity").textContent = result.data.city;
      document.getElementById("profileCountry").textContent = result.data.country;
      document.getElementById("profileImageContainer").src = "." + result.data.profile_image;
    } else {
      document.getElementById("profileUploadError").innerText = get_user_data_by_token_message[result.message];
    }
  });
}

async function uploadFile(type = null) {
  const token = localStorage.getItem("token");
  let fileInput = "";
  if (type == "profile") {
    fileInput = document.getElementById("profileInput");
  } else {
    fileInput = document.getElementById("fileInput");
  }

  const file = fileInput.files[0];
  const upLoadFileErrorMessage = document.getElementById("upLoadFileError");
  const profileUploadErrorMessage = document.getElementById("profileUploadMessage");
  if (!file) {
    if (type == "profile") {
      profileUploadErrorMessage = "no image";
    } else {
      upLoadFileErrorMessage.textContent = "no file";
    }

    return;
  }

  const formData = new FormData();
  formData.append("file", file);
  if (type == "profile") {
    sendFileRequest("POST", `${BASE_URL}/upload_profile`, formData, token, function (result) {
      if (result.success) {
        loadProfile();
        profileUploadErrorMessage.textContent = "Uploading Success!";
        setTimeout(() => (profileUploadErrorMessage.textContent = ""), 3000);
      } else {
        profileUploadErrorMessage.textContent = upload_profile_message[result.message];
      }
    });
  } else {
    sendFileRequest("POST", `${BASE_URL}/upload_file`, formData, token, function (result) {
      if (result.success && result.status == 201) {
        loadUserMedia();
        upLoadFileErrorMessage.textContent = "Uploading Success!";
        setTimeout(() => (upLoadFileErrorMessage.textContent = ""), 3000);
      } else {
        upLoadFileErrorMessage.textContent = upload_file_message[result.message];
        setTimeout(() => (upLoadFileErrorMessage.textContent = ""), 3000);
      }
    });
  }
}

async function loadUserMedia() {
  const token = localStorage.getItem("token");

  sendFileRequest("GET", `${BASE_URL}/get_user_media`, null, token, function (result) {
    if (result.success && result.status == 200) {
      const mediaContainer = document.getElementById("mediaContainer");
      mediaContainer.innerHTML = "";
      result.data.forEach((file) => {
        const filePath = `../${file.file_path.split("/").pop()}`;
        const mediaElement = document.createElement(file.file_type === "mp4" ? "video" : "img");
        mediaElement.src = filePath;
        mediaElement.controls = file.file_type === "mp4";

        mediaContainer.appendChild(mediaElement);
      });
    } else {
      document.getElementById("mediaContainer").innerHTML = "No media found";
    }
  });
}

async function loadProfile() {
  const token = localStorage.getItem("token");

  sendRequest("GET", `${BASE_URL}/get_user_profile`, null, token, function (result) {
    if (result.success && result.status == 200) {
      const profileImg = document.getElementById("profileImageContainer");

      const filePath = `../${result.data.profile_image.split("/").pop()}`;
      profileImg.src = filePath;
    } else {
      document.getElementById("profileUploadError").textContent = load_profile_message[result.message];
    }
  });
}
async function uploadProfile() {
  uploadFile("profile");
}

function delay(ms) {
  return new Promise((resolve) => setTimeout(resolve, ms));
}
async function fetchAddress(lat, lon) {
  try {
    await delay(1100);
    const res = await fetch(`https://geocode.xyz/${lat},${lon}?geoit=json`);
    const data = await res.json();

    if (data.error) {
      return "Unknown location";
    }

    return `${data.city || "Unknown city"}, ${data.country || "Unknown country"}`;
  } catch (e) {
    return "Location lookup failed";
  }
}
